#ifndef __TEST_TYPE
#define __TEST_TYPE

#include <utility>
#include <iostream>

class TestType;

class TestWrapper
{
private:
    const TestType & d_val;
    
public:
    inline TestWrapper(const TestType & val)
        : d_val(val)
    {
    }
    
    void print() const
    {
        std::cout << &d_val;
    }
};

template<class Op, class Data>
class TestExpression;

extern void setZero(TestType & r)
{
    std::cout << &r << " = 0\n";
}

extern void setOne(TestType & r)
{
    std::cout << &r << " = 1\n";
}

extern void add(TestType & r, const TestType & a, const TestType & b)
{
    std::cout << &r << " = " << &a << " + " << &b << "\n";
}

extern void sub(TestType & r, const TestType & a, const TestType & b)
{
    std::cout << &r << " = " << &a << " - " << &b << "\n";
}

extern void mul(TestType & r, const TestType & a, const TestType & b)
{
    std::cout << &r << " = " << &a << " * " << &b << "\n";
}

extern void div(TestType & r, const TestType & a, const TestType & b)
{
    std::cout << &r << " = " << &a << " / " << &b << "\n";
}

extern void mod(TestType & r, const TestType & a, const TestType & b)
{
    std::cout << &r << " = " << &a << " % " << &b << "\n";
}

extern void neg(TestType & r, const TestType & a)
{
    std::cout << &r << " = -" << &a << "\n";
}

class TestType
{
public:
    TestType()
    {
        std::cout << "create " << this << "\n";
    }

    TestType(const TestType & src)
    {
        std::cout << "create " << this << " from " << &src << "\n";
    }

    template<class O, class D>
    TestType(const TestExpression<O, D> & src)
    {
        std::cout << "create " << this << " from ";
        src.print();
        std::cout << "\n";
    }

    ~TestType()
    {
        std::cout << "destroy " << this << "\n";
    }

    TestType & operator = (const TestType & src)
    {
        std::cout << "copy " << this << " from " << &src << "\n";
        return *this;
    }
    
    template<class O, class D>
    TestType & operator = (const TestExpression<O, D> & e)
    {
        std::cout << this << " = ";
        e.print();
        std::cout << "\n";
        return *this;
    }
    
    TestType & operator += (const TestType & b)
    {
        std::cout << this << " += " << &b << "\n";
        return *this;
    }
    
    TestType & operator -= (const TestType & b)
    {
        std::cout << this << " -= " << &b << "\n";
        return *this;
    }
    
    TestType & operator *= (const TestType & b)
    {
        std::cout << this << " *= " << &b << "\n";
        return *this;
    }
    
    TestType & operator /= (const TestType & b)
    {
        std::cout << this << " /= " << &b << "\n";
        return *this;
    }
    
    TestType & operator %= (const TestType & b)
    {
        std::cout << this << " %= " << &b << "\n";
        return *this;
    }
    
    template<class O, class D>
    TestType & operator += (const TestExpression<O, D> & e)
    {
        std::cout << this << " += ";
        e.print();
        std::cout << "\n";
        return *this;
    }
    
    template<class O, class D>
    TestType & operator -= (const TestExpression<O, D> & e)
    {
        std::cout << this << " -= ";
        e.print();
        std::cout << "\n";
        return *this;
    }
    
    template<class O, class D>
    TestType & operator *= (const TestExpression<O, D> & e)
    {
        std::cout << this << " *= ";
        e.print();
        std::cout << "\n";
        return *this;
    }
    
    template<class O, class D>
    TestType & operator /= (const TestExpression<O, D> & e)
    {
        std::cout << this << " /= ";
        e.print();
        std::cout << "\n";
        return *this;
    }
    
    template<class O, class D>
    TestType & operator %= (const TestExpression<O, D> & e)
    {
        std::cout << this << " %= ";
        e.print();
        std::cout << "\n";
        return *this;
    }
};

template<class Op, class Data>
class TestExpression
{
private:
    Op d_op;
    Data d_data;
    
public:
    inline TestExpression(const Op & op, const Data & data)
        : d_op(op), d_data(data)
    {
    }
    
    operator TestType () const
    {
        std::cout << "casting TestExpression[";
        d_op.print(d_data);
        std::cout << "] to TestType()\n";
        return TestType();
    }
    
    void print() const
    {
        d_op.print(d_data);
    }
};

class AddOp
{
public:
    template<class A, class B>
    void print(const std::pair<A, B> & data) const
    {
        std::cout << "(";
        data.first.print();
        std::cout << " + ";
        data.second.print();
        std::cout << ")";
    }
};

class SubOp
{
public:
    template<class A, class B>
    void print(const std::pair<A, B> & data) const
    {
        std::cout << "(";
        data.first.print();
        std::cout << " - ";
        data.second.print();
        std::cout << ")";
    }
};

class MulOp
{
public:
    template<class A, class B>
    void print(const std::pair<A, B> & data) const
    {
        std::cout << "(";
        data.first.print();
        std::cout << " * ";
        data.second.print();
        std::cout << ")";
    }
};

class DivOp
{
public:
    template<class A, class B>
    void print(const std::pair<A, B> & data) const
    {
        std::cout << "(";
        data.first.print();
        std::cout << " / ";
        data.second.print();
        std::cout << ")";
    }
};

class ModOp
{
public:
    template<class A, class B>
    void print(const std::pair<A, B> & data) const
    {
        std::cout << "(";
        data.first.print();
        std::cout << " % ";
        data.second.print();
        std::cout << ")";
    }
};

class NegOp
{
public:
    template<class A>
    void print(const A & data) const
    {
        std::cout << "-";
        data.print();
    }
};

TestExpression<AddOp, std::pair<TestWrapper, TestWrapper> > operator + (const TestType & a, const TestType & b)
{ return TestExpression<AddOp, std::pair<TestWrapper, TestWrapper> >(AddOp(), std::make_pair(TestWrapper(a), TestWrapper(b))); }
TestExpression<SubOp, std::pair<TestWrapper, TestWrapper> > operator - (const TestType & a, const TestType & b)
{ return TestExpression<SubOp, std::pair<TestWrapper, TestWrapper> >(SubOp(), std::make_pair(TestWrapper(a), TestWrapper(b))); }
TestExpression<MulOp, std::pair<TestWrapper, TestWrapper> > operator * (const TestType & a, const TestType & b)
{ return TestExpression<MulOp, std::pair<TestWrapper, TestWrapper> >(MulOp(), std::make_pair(TestWrapper(a), TestWrapper(b))); }
TestExpression<DivOp, std::pair<TestWrapper, TestWrapper> > operator / (const TestType & a, const TestType & b)
{ return TestExpression<DivOp, std::pair<TestWrapper, TestWrapper> >(DivOp(), std::make_pair(TestWrapper(a), TestWrapper(b))); }
TestExpression<ModOp, std::pair<TestWrapper, TestWrapper> > operator % (const TestType & a, const TestType & b)
{ return TestExpression<ModOp, std::pair<TestWrapper, TestWrapper> >(ModOp(), std::make_pair(TestWrapper(a), TestWrapper(b))); }

template<class O2, class D2>
TestExpression<AddOp, std::pair<TestWrapper, TestExpression<O2, D2> > > operator + (const TestType & a, const TestExpression<O2, D2> & b)
{ return TestExpression<AddOp, std::pair<TestWrapper, TestExpression<O2, D2> > >(AddOp(), std::make_pair(TestWrapper(a), b)); }
template<class O2, class D2>
TestExpression<SubOp, std::pair<TestWrapper, TestExpression<O2, D2> > > operator - (const TestType & a, const TestExpression<O2, D2> & b)
{ return TestExpression<SubOp, std::pair<TestWrapper, TestExpression<O2, D2> > >(SubOp(), std::make_pair(TestWrapper(a), b)); }
template<class O2, class D2>
TestExpression<MulOp, std::pair<TestWrapper, TestExpression<O2, D2> > > operator * (const TestType & a, const TestExpression<O2, D2> & b)
{ return TestExpression<MulOp, std::pair<TestWrapper, TestExpression<O2, D2> > >(MulOp(), std::make_pair(TestWrapper(a), b)); }
template<class O2, class D2>
TestExpression<DivOp, std::pair<TestWrapper, TestExpression<O2, D2> > > operator / (const TestType & a, const TestExpression<O2, D2> & b)
{ return TestExpression<DivOp, std::pair<TestWrapper, TestExpression<O2, D2> > >(DivOp(), std::make_pair(TestWrapper(a), b)); }
template<class O2, class D2>
TestExpression<ModOp, std::pair<TestWrapper, TestExpression<O2, D2> > > operator % (const TestType & a, const TestExpression<O2, D2> & b)
{ return TestExpression<ModOp, std::pair<TestWrapper, TestExpression<O2, D2> > >(ModOp(), std::make_pair(TestWrapper(a), b)); }

template<class O1, class D1>
TestExpression<AddOp, std::pair<TestExpression<O1, D1>, TestWrapper> > operator + (const TestExpression<O1, D1> & a, const TestType & b)
{ return TestExpression<AddOp, std::pair<TestExpression<O1, D1>, TestWrapper> >(AddOp(), std::make_pair(a, TestWrapper(b))); }
template<class O1, class D1>
TestExpression<SubOp, std::pair<TestExpression<O1, D1>, TestWrapper> > operator - (const TestExpression<O1, D1> & a, const TestType & b)
{ return TestExpression<SubOp, std::pair<TestExpression<O1, D1>, TestWrapper> >(SubOp(), std::make_pair(a, TestWrapper(b))); }
template<class O1, class D1>
TestExpression<MulOp, std::pair<TestExpression<O1, D1>, TestWrapper> > operator * (const TestExpression<O1, D1> & a, const TestType & b)
{ return TestExpression<MulOp, std::pair<TestExpression<O1, D1>, TestWrapper> >(MulOp(), std::make_pair(a, TestWrapper(b))); }
template<class O1, class D1>
TestExpression<DivOp, std::pair<TestExpression<O1, D1>, TestWrapper> > operator / (const TestExpression<O1, D1> & a, const TestType & b)
{ return TestExpression<DivOp, std::pair<TestExpression<O1, D1>, TestWrapper> >(DivOp(), std::make_pair(a, TestWrapper(b))); }
template<class O1, class D1>
TestExpression<ModOp, std::pair<TestExpression<O1, D1>, TestWrapper> > operator % (const TestExpression<O1, D1> & a, const TestType & b)
{ return TestExpression<ModOp, std::pair<TestExpression<O1, D1>, TestWrapper> >(ModOp(), std::make_pair(a, TestWrapper(b))); }

template<class O1, class D1, class O2, class D2>
TestExpression<AddOp, std::pair<TestExpression<O1, D1>, TestExpression<O2, D2> > > operator + (const TestExpression<O1, D1> & a, const TestExpression<O2, D2> & b)
{ return TestExpression<AddOp, std::pair<TestExpression<O1, D1>, TestExpression<O2, D2> > >(AddOp(), std::make_pair(a, b)); }
template<class O1, class D1, class O2, class D2>
TestExpression<SubOp, std::pair<TestExpression<O1, D1>, TestExpression<O2, D2> > > operator - (const TestExpression<O1, D1> & a, const TestExpression<O2, D2> & b)
{ return TestExpression<SubOp, std::pair<TestExpression<O1, D1>, TestExpression<O2, D2> > >(SubOp(), std::make_pair(a, b)); }
template<class O1, class D1, class O2, class D2>
TestExpression<MulOp, std::pair<TestExpression<O1, D1>, TestExpression<O2, D2> > > operator * (const TestExpression<O1, D1> & a, const TestExpression<O2, D2> & b)
{ return TestExpression<MulOp, std::pair<TestExpression<O1, D1>, TestExpression<O2, D2> > >(MulOp(), std::make_pair(a, b)); }
template<class O1, class D1, class O2, class D2>
TestExpression<DivOp, std::pair<TestExpression<O1, D1>, TestExpression<O2, D2> > > operator / (const TestExpression<O1, D1> & a, const TestExpression<O2, D2> & b)
{ return TestExpression<DivOp, std::pair<TestExpression<O1, D1>, TestExpression<O2, D2> > >(DivOp(), std::make_pair(a, b)); }
template<class O1, class D1, class O2, class D2>
TestExpression<ModOp, std::pair<TestExpression<O1, D1>, TestExpression<O2, D2> > > operator % (const TestExpression<O1, D1> & a, const TestExpression<O2, D2> & b)
{ return TestExpression<ModOp, std::pair<TestExpression<O1, D1>, TestExpression<O2, D2> > >(ModOp(), std::make_pair(a, b)); }

TestExpression<NegOp, TestWrapper> operator - (const TestType & a)
{ return TestExpression<NegOp, TestWrapper>(NegOp(), TestWrapper(a)); }

template<class O1, class D1>
TestExpression<NegOp, TestExpression<O1, D1> > operator - (const TestExpression<O1, D1> & a)
{ return TestExpression<NegOp, TestExpression<O1, D1> >(NegOp(), a); }

#endif
